from abc import ABC
from abc import abstractmethod
from collections import Iterable
from io import StringIO

from concentrator import settings
from kinder.core.declaration_status.enum import DSS
from kinder.core.unit.models import Unit
from kinder.core.unit.models import UnitKind


class SMEV3Response:

    def __init__(self, response, logging_data):
        """Настройка ответа.

        :param response: ответ
        :type response: requests.Response
        :param logging_data: дополнительные данные для логирования
        :type logging_data: dict
        """

        self.response = response
        self.logging_data = logging_data


class BaseExecutor(ABC):
    """Базовый класс исполнителя сервиса СМЭВ 3."""

    name_service = None
    type_service = None

    @classmethod
    def __subclasshook__(cls, subclass):

        if cls is BaseExecutor:
            return any(
                True for sub in subclass.__mro__
                if sub.__dict__.get('process') is not None)

        return NotImplemented

    @classmethod
    @abstractmethod
    def process(cls, message, request_body):
        """Запускает выполнение логики обработки сервиса.

        :param message: xml пакет
        :param request_body: тип

        :return: инстанс ответа
        :rtype: SMEV3Response
        """
        pass


class SMEV3RepositoryExecutors:
    """Хранилище исполнителей сервисов СМЭВ 3."""

    _store = [
        # (Наименование сервиса, Тип сервиса, Обработчик сервиса),
        # ...
    ]

    @classmethod
    def set_up(cls, config):
        """Инициализирует хранилище класса.

        Принимает список наследников класса BaseExecutor
        или структуру вида:
        (Наименование сервиса, Тип сервиса, Обработчик сервиса).

        :param config: список настроек
        :type config: Union[
            Iterable[Tuple[str, Type, callable]],
            Iterable[BaseExecutor]]
        :return: None
        """
        cls._store.extend(
            (i.name_service, i.type_service, i.process)
            if not isinstance(i, Iterable) and issubclass(i, BaseExecutor)
            else i for i in config)

    @classmethod
    def get_executor(cls, message):
        """Определяет исполнителя для полученного пакета.

        :param message: Тип, содержащий данные сервиса
        :return: Ссылку на метод исполнителя
        """

        for name_service, type_service, executor in cls._store:
            service = getattr(message, name_service, None)
            if (service
                    and isinstance(service, type_service)
                    and callable(executor)):
                return executor


def render_type2xml(type_, name_type, pretty_print=False):
    """Рендерит классы в xml.

    :param type_: инстанс типа
    :param name_type: наименование типа
    :param pretty_print: признак "красивого" вывода в xml
    :return: сформированный xml
    """

    buffer = StringIO()
    type_.export(
        buffer, 0, name_=name_type, pretty_print=pretty_print)

    return buffer.getvalue()


def is_cancel_allowed(status):
    """Проверяет разрешена ли отмена заявления в указанном статусе

    :param status: Статус заявления
    :type status: DeclarationStatus
    :return: Разрешена ли отмена
    :rtype: bool
    """

    if status.code in list(DSS.values.keys()):
        if getattr(settings, status.code.upper(), False):
            return True
    elif settings.OTHER_STATUS:
        return True

    return False


def get_oktmo_region():
    """Возвращает ОКТМО головного учреждения.

    :return: октмо или None
    :rtype: Optional[str]
    """

    region = Unit.objects.filter(kind=UnitKind.REGION).last()

    return region.octmo if region else None

