from aio_client.provider.api import PostProviderRequest
from aio_client.provider.api import push_request

from concentrator.smev3_v321.base.constants import SUCCESS_MESSAGE
from concentrator.smev3_v321.base.exceptions import ContentFailure
from concentrator.smev3_v321.base.utils import BaseExecutor
from concentrator.smev3_v321.base.utils import SMEV3Response
from concentrator.smev3_v321.base.utils import is_cancel_allowed
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.service_types import kinder_conc
from kinder.core.declaration.models import Declaration
from kinder.core.declaration_status.enum import DSS
from kinder.core.declaration_status.models import DeclarationStatus
from kinder.webservice.api.declaration import get_declaration_by_client_id
from kinder.webservice.api.exceptions import ApiException

from .constants import CANCEL_SUCCESS_COMMENT
from .constants import DECL_CANT_REJECTED
from .constants import DECL_NOT_FOUND
from .constants import WHY_CHANGE


class CancelRequestExecutor(BaseExecutor):
    """Исполнитель сервиса cancelRequest."""

    name_service = 'cancelRequest'
    type_service = kinder_conc.cancelRequestType

    @classmethod
    def process(cls, message, request_body):
        """
        Обработка запроса

        :param message: Параметры запроса
        :type message: dict
        :param request_body: Объект FormDataType
        :type request_body: concentrator.smev3.service_types.kinder_conc
        .FormDataType
        :return: Ответ
        :rtype: SMEV3Response
        """

        try:
            response_body = render_type2xml(
                cls.get_response(request_body),
                name_type='FormDataResponse'
            )
            content_failure_code = None
            content_failure_comment = None

        except ContentFailure as exc:
            response_body = render_type2xml(
                kinder_conc.FormDataResponseType(
                    cancelResponse=kinder_conc.cancelResponseType(
                        orderId=int(request_body.cancelRequest.orderId),
                        result=exc.content_failure_code,
                        comment=exc.content_failure_comment
                    )
                ), name_type='FormDataResponse')
            content_failure_code = exc.content_failure_code
            content_failure_comment = exc.content_failure_comment

        response = push_request(PostProviderRequest(
            origin_message_id=message['origin_message_id'],
            body=response_body,
            message_type=message['message_type'],
            attachments=None,
            content_failure_code=content_failure_code,
            content_failure_comment=content_failure_comment or '',
            replay_to=message['replay_to'],
            is_test_message=message['is_test_message']
        ))

        return SMEV3Response(
            response,
            {
                'method_name': cls.name_service,
                'response': response_body,
                'result': content_failure_comment or SUCCESS_MESSAGE
            })

    @classmethod
    def get_response(cls, request_body):
        """Формирует ответ на cancelRequest.

        :param request_body: Запрос на отмену заявления
        :type request_body: FormDataType

        :return: Ответ на запрос cancelRequest
        :rtype: FormDataResponseType
        :raise: ContentFailure
        """

        try:
            declaration = get_declaration_by_client_id(
                Declaration.objects.select_related('status'),
                request_body.cancelRequest.orderId
            )
        except ApiException:
            declaration = None

        if not declaration:
            raise ContentFailure(
                kinder_conc.CancelResultType.REJECTED, DECL_NOT_FOUND)
        elif (declaration.status.code in
              [DSS.no_active_statuses(), DSS.DIRECTED] or
              not is_cancel_allowed(declaration.status)):
            raise ContentFailure(
                kinder_conc.CancelResultType.REJECTED, DECL_CANT_REJECTED)

        try:
            declaration.change_status(DeclarationStatus.objects.get(
                code=DSS.REFUSED), why_change=WHY_CHANGE)
        except DeclarationStatus.DoesNotExist:
            raise ContentFailure(
                kinder_conc.CancelResultType.REJECTED,
                f'Статуса {DSS.values.get(DSS.REFUSED)} не существует'
            )

        response = kinder_conc.FormDataResponseType(
            cancelResponse=kinder_conc.cancelResponseType(
                orderId=int(request_body.cancelRequest.orderId),
                result=kinder_conc.CancelResultType.CANCELLED,
                comment=CANCEL_SUCCESS_COMMENT)
        )

        return response
