from collections import namedtuple

from django.db.models import Q

from concentrator.smev3_v321.esnsi.enums import DataTypesEnum as DT
from concentrator.smev3_v321.esnsi.settings import UPDATE_CLASSIFIER_REGOKATO
from kinder.core.dict.models import UnitKind
from kinder.core.unit.models import Unit


class ClassifierBase:
    """Базовый класс справочников для передачи в ЕСНСИ (СМЭВ 3)."""

    model = None

    # имя и тип атрибута записи для ключевого поля классификатора
    pk_attr_name = None
    pk_attr_type = None

    def __init__(self):
        assert self.model

        self._filename = 'update'
        self._storage_sections = (
            'update_esnsi',
            self.get_service_code(),
        )

    def get_pk_val(self, obj):
        return obj.id

    def data_queryset(self):
        return self.model.objects.all()

    def get_obj(self, row):
        """Возвращает объект элемента выборки"""

        # если в исходном запросе указан values, то выдает объект с
        # отображенными в его атрибуты ключами словаря,
        # иначе отдается исходный элемент выборки
        res = row
        if isinstance(row, dict):
            res = namedtuple('row_val', row)(**row)
        return res

    def queryset_filter(self):
        """Возвращает актуальный QuerySet выборки данных для выгрузки."""
        return Q()

    @classmethod
    def get_service_code(cls):
        """Получить имя классификатора."""
        from .models import ESNSIClassifier
        return ESNSIClassifier.objects.get(classifier_class=cls.__name__).code


class UnitClassifier(ClassifierBase):
    """Организации (детсады)"""

    model = Unit

    def queryset_filter(self):
        return Q(kind=UnitKind.DOU)

    def get_attrs(self, obj):
        return {
            'CODE': (DT.STRING, (f'{obj.id}' or '')[:50]),
            'TITLE': (DT.STRING, (obj.name or '')[:255]),
            'REGOKATO': (DT.STRING,UPDATE_CLASSIFIER_REGOKATO[:10]),
            'ADDRESS': (DT.STRING, (obj.address_full or '')[:255]),
            'FIAS': (DT.STRING, (obj.address_place or '')[:50]),
            'PHONE': (DT.STRING, (obj.telephone or '')[:50]),
            'EMAIL': (DT.STRING, (obj.email or '')[:50]),
            'SCHEDULE': (DT.STRING, (obj.reception_time or '')[:100])
        }


class MaxDooClassifier(ClassifierBase):
    """Максимальное количество детсадов, которое может выбрать заявитель"""

    # Пуск - Организации - МО - поле "Максимальное количество желаемых ДОО"
    model = Unit

    def queryset_filter(self):
        return Q(kind=UnitKind.MO)

    def get_attrs(self, obj):
        code = obj.address_place or f'{obj.id}'
        return {
            'CODE': (DT.STRING, (code or '')[:50]),
            'TITLE': (DT.STRING, (obj.name or '')[:255]),
            'REGOKATO': (DT.STRING, UPDATE_CLASSIFIER_REGOKATO[:10]),
            'EDUORGMAX': (DT.INTEGER, (
                    obj.max_desired_dou if obj.max_desired_dou in range(11)
                    else '')),
        }
