from django.db import transaction
from m3.plugins import ExtensionManager

from concentrator.smev3_v321.esnsi.classifiers import MaxDooClassifier
from concentrator.smev3_v321.esnsi.classifiers import UnitClassifier
from concentrator.smev3_v321.esnsi.settings import UPDATE_CLASSIFIER_REGOKATO
from concentrator.smev3_v321.esnsi.tasks import EsnsiSendingTask
from concentrator.smev3_v321.models import ESNSIClassifier
from concentrator.smev3_v321.models import UpdateClassifierRequest
from kinder.core.dict.models import GroupTypeEnumerate
from kinder.helpers import RequestFactory


class ESNSIRequestFactory(RequestFactory):
    """Фабрика запросов ЕСНСИ
    Обязательные параметры:
    classifier - классификатор ЕСНСИ
    request_type - Тип запроса ЕСНСИ
    instances - один или несколько инстансов соответствующей классификатору модели
    """

    def __init__(self, classifier=None, request_type=None, instances=None):
        super().__init__(instances)
        self.classifier = classifier
        self.request_type = request_type

    def _collect_data_part(self, instance, code_prefix=None):
        code = instance.id
        if code_prefix:
            code = f'{code_prefix}_{code}'
        return {
            'CODE': code,
            'TITLE': instance.name,
            'REGOKATO': UPDATE_CLASSIFIER_REGOKATO
        }

    def create_request(self):
        """Создание запроса."""

        data = self._collect_data()

        request = UpdateClassifierRequest.objects.create(
            request_type=UpdateClassifierRequest.UPDATE_CLASSIFIERS,
            classifier=self.classifier,
            data=data
        )
        return request


class ESNSITaskFactory(ESNSIRequestFactory):
    """Фабрика задач ЕСНСИ"""

    def _create_task(self):
        """Создание задачи"""

        return EsnsiSendingTask()

    def set_task(self):
        """Запуск задачи"""

        task = self._create_task()
        request_id = self.create_request().id
        return transaction.on_commit(lambda: task.apply_async((request_id,)))


class ESNSIDelegateTaskFactory(ESNSITaskFactory):
    """Фабрика задач ЕСНСИ для справочника Виды документов, подтверждающих
    право представлять интересы ребенка."""

    def _collect_first_data_part(self, instance):
        data = self._collect_data_part(instance, code_prefix='op')
        data.update({'PERSONTYPE': 3})
        return data

    def _collect_second_data_part(self, instance):
        data = self._collect_data_part(instance, code_prefix='us')
        data.update({'PERSONTYPE': 4})
        return data

    def _collect_data(self):
        """Сбор данных запроса."""

        data = []
        for instance in self.instances:
            data.append(self._collect_first_data_part(instance))
            data.append(self._collect_second_data_part(instance))
        return data


class ESNSIUnitTaskFactory(ESNSITaskFactory):
    """Фабрика задач ЕСНСИ для справочника Виды документов, подтверждающих
    право представлять интересы ребенка."""

    def _collect_data_part(self, instance):
        data = super()._collect_data_part(instance)
        data.update({
            'ADDRESS': instance.address_full,
            'FIAS': instance.address_place,
            'PHONE': instance.telephone,
            'EMAIL': instance.email,
            'SCHEDULE': instance.reception_time
        })
        return data


class ESNSIGroupTaskFactory(ESNSITaskFactory):
    """Фабрика задач ЕСНСИ для справочника Виды документов, подтверждающих
    право представлять интересы ребенка."""

    def _collect_data_part(self, instance):
        part = super()._collect_data_part(instance)
        if instance.code == GroupTypeEnumerate.COMP:
            part.update({'ADAPTATIONGROUP': '2'})
        elif instance.code == GroupTypeEnumerate.HEALTH:
            part.update({'ADAPTATIONGROUP': '3'})
        return part


class ESNSITaskFacade:
    """Фасад для связывания Фабрик задач ЕСНСИ с соответствующими
    классификаторами."""

    CLASSIFIERS = {
        UnitClassifier.__name__: ESNSIUnitTaskFactory,
        MaxDooClassifier.__name__: ESNSIUnitTaskFactory,
    }

    def __init__(self, classifier, request_type, instances):
        self._classifier = classifier
        self._request_type = request_type
        self._instances = instances
        self._extend_classifiers()

    def _chose_factory(self):
        """Выбор соответствующей классификатору фабрики"""

        return self.CLASSIFIERS.get(self._classifier.classifier_class)

    def chose_factory_set_task(self):
        """Запуск таски для выбранной фабрики"""

        factory = self._chose_factory()
        if not factory:
            return
        return factory(
            self._classifier,
            self._request_type,
            self._instances
        ).set_task()

    def _extend_classifiers(self):
        """Расширение для добавления классификаторов"""

        extra_classifiers = ExtensionManager().execute(
            'esnsi_classifiers_facade_extension')
        if extra_classifiers:
            self.CLASSIFIERS.update(extra_classifiers)


def get_classifier_instance(classifier):
    """Получение инстанса модели ESNSIClassifier по классификатору"""
    return ESNSIClassifier.objects.get(classifier_class=classifier.__name__)
