from django.contrib.postgres.fields import JSONField
from django.db import models
from m3.db import BaseEnumerate
from m3.plugins import ExtensionManager

from concentrator.smev3_v321.esnsi.classifiers import MaxDooClassifier
from concentrator.smev3_v321.esnsi.classifiers import UnitClassifier
from kinder.core.audit_log.managers import AuditLog
from kinder.core.direct.models import Direct
from kinder.core.models import LogableModel
from kinder.webservice.smev3.models.consumer import BaseConsumerSmevRequest


class ESNSIClassifier(LogableModel):
    """Справочники для передачи в ЕСНСИ"""

    audit_log = AuditLog()

    classifiers_classes = (
        UnitClassifier,
        MaxDooClassifier,
    )

    classifiers_classes_choices = (
        (UnitClassifier.__name__, 'Организации (детсады)'),
        (MaxDooClassifier.__name__,
         'Максимальное количество детсадов, которое может выбрать заявитель'
         ),
    )

    uid = models.CharField(
        'Уникальный идентификатор классификатора', max_length=36,
        null=True, blank=True
    )

    code = models.CharField(
        'Код классификатора', max_length=100, null=True, blank=True
    )

    name = models.CharField(
        'Наименование справочника', max_length=250, default='',
    )

    classifier_class = models.CharField(
        'Класс классификатора', max_length=50, default='', unique=True,
        choices=classifiers_classes_choices
    )

    class Meta:
        verbose_name = 'Справочник для передачи в ЕСНСИ (СМЭВ 3)'
        verbose_name_plural = 'Справочник для передачи в ЕСНСИ (СМЭВ 3)'
        app_label = 'smev3_v321'
        db_table = 'concentrator_smev3_v321_esnsiclassifier'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._extend_classifiers()

    def get_model_by_classifier_name(self, classifier_name):
        for cl in self.classifiers_classes:
            if cl.__name__ == classifier_name:
                return cl.model

    def _extend_classifiers(self):
        """Расширение для добавления классификаторов"""

        extra_classifier = ExtensionManager().execute(
            'esnsi_classifiers_model_extension')
        if extra_classifier:
            self.classifiers_classes += (extra_classifier,)


class UpdateClassifierRequest(BaseConsumerSmevRequest):
    """
    Запрос отправки справочника в ЕСНСИ
    """

    audit_log = AuditLog()

    RESULT_SUCCESS = 1
    RESULT_FAILURE = 2

    RESULT_CHOICES = (
        (RESULT_SUCCESS, 'Успешно'),
        (RESULT_FAILURE, 'Ошибка'),
    )

    UPDATE_CLASSIFIERS = 'update'
    DELETE_CLASSIFIERS = 'delete'

    REQUEST_TYPE_CHOICES = (
        (UPDATE_CLASSIFIERS, 'Обновление данных справочника'),
        (DELETE_CLASSIFIERS, 'Удаление данных справочника'),
    )

    request_result = models.PositiveSmallIntegerField(
        verbose_name='Результат',
        choices=RESULT_CHOICES,
        null=True, blank=True
    )

    request_sent = models.DateTimeField(
        verbose_name='Дата и время запроса',
        null=True, blank=True
    )

    response_returned = models.DateTimeField(
        verbose_name='Дата и время ответа',
        null=True, blank=True
    )

    request_type = models.CharField(
        verbose_name='Тип запроса',
        max_length=32,
        choices=REQUEST_TYPE_CHOICES
    )
    classifier = models.ForeignKey(
        'smev3_v321.ESNSIClassifier',
        blank=True, null=True,
        on_delete=models.CASCADE
    )

    data = JSONField(
        verbose_name='Данные запроса',
        null=False, blank=False
    )

    class Meta:
        verbose_name = 'Запрос на обновление данных в ЕСНСИ (СМЭВ 3)'
        verbose_name_plural = 'Запросы на обновление данных в ЕСНСИ (СМЭВ 3)'
        app_label = 'smev3_v321'
        db_table = 'concentrator_smev3_v321_updateclassifierrequest'


class ApplicantAnswersEnum(BaseEnumerate):
    """Значения поля "Ответ заявителя"."""

    values = {
        True: 'Согласен',
        False: 'Отказался',
    }

    # Словарь со значениями для использования в фильтре в реестре направлений
    values_for_filter = values.copy()
    values_for_filter['empty'] = 'Пусто'


class ApplicantAnswer(LogableModel):
    """Ответ заявителя для направления."""

    direct = models.OneToOneField(
        Direct, on_delete=models.CASCADE,
        related_name='smev3_v321_applicant_answer',
        verbose_name='Направление'
    )

    answer = models.BooleanField(verbose_name='Ответ заявителя')

    comment = models.CharField(
        max_length=2048, null=True, blank=True,
        verbose_name='Комментарий заявителя',
    )

    audit_log = AuditLog()

    class Meta:
        verbose_name = 'Ответ заявителя'
        db_table = 'concentrator_smev3_v321_applicantanswer'


class AttachmentRequest(BaseConsumerSmevRequest):
    """
    Запросы по взаимодействию с ВС 
    "Получение списков детей, получивших места в дошкольных организациях"
    """
